#include "stdafx.h"
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <direct.h>
#include <errno.h>
#include <math.h>
#include <time.h>
#include <process.h>
#include <cuda.h>

#define h 480
#define w 720
#define THREADX 16
#define THREADY 8

texture<float, 2, cudaReadModeElementType> tex;

__global__ void sobelGrad_rowFilter(float *pass1_out,int ww, int hh, float rMask[3])
{
	__shared__ float smImage[THREADY][THREADX+2];
	__shared__ float smOut[THREADY][THREADX];	

	//Block Index
	unsigned char bx=blockIdx.x;
	unsigned char by=blockIdx.y;
	   
	//Thread Index
	unsigned char tx=threadIdx.x;
	unsigned char ty=threadIdx.y;

	//Row and Column Index of Global Grid
	unsigned short int row = by * THREADY + ty;
	unsigned short int col = bx * THREADX + tx;

	int global_index = row*ww + col;
	int prev_block_col_index = THREADX*bx-1;
	int next_block_col_index = THREADX*(bx+1);

	//Load Center
	smImage[ty][1+tx] = tex2D(tex,col,row);	

	//Load Left Apron
	smImage[ty][0] = tex2D(tex,prev_block_col_index,row);

	//Load Right Apron
	smImage[ty][THREADX+1] = tex2D(tex,next_block_col_index,row);

	__syncthreads();

	smOut[ty][tx] = smImage[ty][tx]*rMask[0]+smImage[ty][tx+1]*rMask[1]+smImage[ty][tx+2]*rMask[2];

	__syncthreads();

	pass1_out[global_index] = smOut[ty][tx];
}

__global__ void sobelGrad_colFilter(float *in,float *GradOut, int ww, int hh, float cMask[3])
{
	__shared__ float smImage[THREADY+2][THREADX];
	__shared__ float smOut[THREADY][THREADX];	

	//Block Index
	unsigned char bx=blockIdx.x;
	unsigned char by=blockIdx.y;
	   
	//Thread Index
	unsigned char tx=threadIdx.x;
	unsigned char ty=threadIdx.y;

	//Row and Column Index of Global Grid
	unsigned short int row = by * THREADY + ty;
	unsigned short int col = bx * THREADX + tx;

	int global_index = row*ww + col;
	int prev_block_row_index = THREADY*by-1;
	int next_block_row_index = THREADY*(by+1);

	//Load Center
	smImage[1+ty][tx] = in[global_index];

	//Load Top Apron
	smImage[0][tx] = in[prev_block_row_index*ww + col];

	//Load Bottom Apron
	smImage[THREADY+1][tx] = in[next_block_row_index*ww + col];

	__syncthreads();

	smOut[ty][tx] = smImage[ty][tx]*cMask[0]+smImage[ty+1][tx]*cMask[1]+smImage[ty+2][tx]*cMask[2];
	
	__syncthreads();

	GradOut[global_index] = smOut[ty][tx];
}

__global__ void find_edge(float *GradX, float *GradY, float *Out, int ww, int hh)
{	
	float gx,gy,result;

	//Block Index
	unsigned char bx=blockIdx.x;
	unsigned char by=blockIdx.y;
	   
	//Thread Index
	unsigned char tx=threadIdx.x;
	unsigned char ty=threadIdx.y;

	//Row and Column Index of Global Grid
	unsigned short int row = by * THREADY + ty;
	unsigned short int col = bx * THREADX + tx;

	int global_index = row*ww + col;

	gx = GradX[global_index];
	gy = GradY[global_index];

	result = abs(int(sqrt(double(gx*gx+gy*gy))));

	if(result>85)
		result=255;
	else
		result=0;

	Out[global_index] = result;
}

void main()
{
	// CPU variables
	FILE *ip,*dp;
	char *cur_file;	
	int i,j;
	float *image, *val, *bin_image;
	float k1[3]={-1,0,1}, k2[3]={1,2,1};

	//GPU variables
	float *xpass1, *ypass1, *gradx, *grady, *edge_out, *mask1, *mask2;

	image=(float*)calloc(h*w,sizeof(float));
	val=(float*)calloc(h*w,sizeof(float));
	bin_image=(float*)calloc(h*w,sizeof(float));
	size_t size=h*w*sizeof(float);
	size_t masksize=3*sizeof(float);

	// Getting pixel values of image: START
	cur_file="D:\\CUDA_Testing\\dasdsa\\03266.bmp";
	ip=fopen(cur_file,"rb");
	fseek(ip,54,SEEK_CUR);
	
	for(i=0;i<h;i++)
	{
		for(j=0;j<w;j++)
		{
			image[i*w+j]=float(floor((0.3*(int)getc(ip))+(0.59*(int)getc(ip))+(0.11*(int)getc(ip))));
		}
	}

	for(i=h-1;i>=0;i--)
	{
		for(j=0;j<w;j++)
		{
			val[(h-1-i)*w+j]=image[i*w+j];
		}
	}
	// Getting pixel values of image: END

	// Sobel Edge Detection on GPU: START
	
	//Declare 2D CUDA Array
	cudaSetDevice(1);
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
	cudaArray* cu_array;

	// set texture parameters
	tex.addressMode[0] = cudaAddressModeClamp;
	tex.addressMode[1] = cudaAddressModeClamp;
	tex.filterMode = cudaFilterModePoint;
	tex.normalized = 0;

	// Allocate Global Memory on GPU
	cudaMalloc((void **) &xpass1, size);
	cudaMalloc((void **) &ypass1, size);
	cudaMalloc((void **) &gradx, size);
	cudaMalloc((void **) &grady, size);
	cudaMalloc((void **) &edge_out, size);
	cudaMalloc((void **) &mask1, masksize);
	cudaMalloc((void **) &mask2, masksize);
	cudaMallocArray( &cu_array, &channelDesc, w, h);

	// CPU to GPU Data Transfer
	cudaMemcpyToArray(cu_array, 0, 0, val, size, cudaMemcpyHostToDevice);
	cudaMemcpy(mask1, k1, masksize, cudaMemcpyHostToDevice);
	cudaMemcpy(mask2, k2, masksize, cudaMemcpyHostToDevice);

	// Bind the array to the texture
	cudaBindTextureToArray(tex, cu_array);
	
	// Define number of threads in a block i.e., block size and number of blocks in a grid i.e., grid size
	dim3 dimBlock(THREADX,THREADY);
	dim3 dimGrid(w/THREADX,h/THREADY);

	// Calculate Gradient in X-direction
	sobelGrad_rowFilter <<< dimGrid, dimBlock >>> (xpass1,w,h,mask1);
	sobelGrad_colFilter <<< dimGrid, dimBlock >>> (xpass1,gradx,w,h,mask2);
	

	// Calculate Gradient in Y-direction
	sobelGrad_rowFilter <<< dimGrid, dimBlock >>> (ypass1,w,h,mask2);
	sobelGrad_colFilter <<< dimGrid, dimBlock >>> (ypass1,grady,w,h,mask1);	

	// Calculate Sobel Edge
	find_edge <<< dimGrid, dimBlock >>> (gradx,grady,edge_out,w,h);	

	// Copy results back to CPU
	cudaMemcpy(bin_image, edge_out, size, cudaMemcpyDeviceToHost);
	
	cudaFree(gradx);
	cudaFree(grady);
	cudaFree(xpass1);
	cudaFree(ypass1);
	cudaFree(mask1);
	cudaFree(mask2);
	cudaFree(edge_out);	
	
	// Sobel Edge Detection on GPU: END
	
	dp=fopen(".\\edge.pgm","wb");
	fprintf(dp, "P5\n%d %d\n255\n", w, h);
	for(i=0;i<h;i++)
		for(j=0;j<w;j++)
			fprintf(dp,"%c",(unsigned char)bin_image[i*w+j]);
}