#include "stdafx.h"
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <direct.h>
#include <errno.h>
#include <math.h>
#include <time.h>
#include <process.h>
#include <cuda.h>

#define h 480
#define ww 720
#define block_size_x 8
#define block_size_y 16

texture<BYTE, 2, cudaReadModeElementType> tex;

__global__ void find_edge(uchar4 *outEdge, int www, int hh)
{
	uchar4 result;
	int4 sx,sy,sl;

	//Block Index
	int bx=blockIdx.x;
	int by=blockIdx.y;
	   
	//Thread Index
	int tx=threadIdx.x;
	int ty=threadIdx.y;

	//row and Column Index of Global Grid
	int row=by*block_size_y+ty;
	int col=bx*block_size_x+tx;

	int global_index = row*www/4 + col;	
	
	sx.x = tex2D(tex,col*4-1,row-1)+2*tex2D(tex,col*4,row-1)+tex2D(tex,col*4+1,row-1)-(tex2D(tex,col*4-1,row+1)+2*tex2D(tex,col*4,row+1)+tex2D(tex,col*4+1,row+1));
	sy.x = tex2D(tex,col*4-1,row-1)+2*tex2D(tex,col*4,row)+tex2D(tex,col*4+1,row+1)-(tex2D(tex,col*4+1,row-1)+2*tex2D(tex,col*4+1,row)+tex2D(tex,col*4+1,row+1));
	
	sx.y = tex2D(tex,col*4+1-1,row-1)+2*tex2D(tex,col*4+1,row-1)+tex2D(tex,col*4+1+1,row-1)-(tex2D(tex,col*4+1-1,row+1)+2*tex2D(tex,col*4+1,row+1)+tex2D(tex,col*4+1+1,row+1));
	sy.y = tex2D(tex,col*4+1-1,row-1)+2*tex2D(tex,col*4+1,row)+tex2D(tex,col*4+1+1,row+1)-(tex2D(tex,col*4+1+1,row-1)+2*tex2D(tex,col*4+1+1,row)+tex2D(tex,col*4+1+1,row+1));
	
	sx.z = tex2D(tex,col*4+2-1,row-1)+2*tex2D(tex,col*4+2,row-1)+tex2D(tex,col*4+2+1,row-1)-(tex2D(tex,col*4+2-1,row+1)+2*tex2D(tex,col*4+2,row+1)+tex2D(tex,col*4+2+1,row+1));
	sy.z = tex2D(tex,col*4+2-1,row-1)+2*tex2D(tex,col*4+2,row)+tex2D(tex,col*4+2+1,row+1)-(tex2D(tex,col*4+2+1,row-1)+2*tex2D(tex,col*4+2+1,row)+tex2D(tex,col*4+2+1,row+1));
	
	sx.w = tex2D(tex,col*4+3-1,row-1)+2*tex2D(tex,col*4+3,row-1)+tex2D(tex,col*4+3+1,row-1)-(tex2D(tex,col*4+3-1,row+1)+2*tex2D(tex,col*4+3,row+1)+tex2D(tex,col*4+3+1,row+1));
	sy.w = tex2D(tex,col*4+3-1,row-1)+2*tex2D(tex,col*4+3,row)+tex2D(tex,col*4+3+1,row+1)-(tex2D(tex,col*4+3+1,row-1)+2*tex2D(tex,col*4+3+1,row)+tex2D(tex,col*4+3+1,row+1));
	
	sl.x=abs(int(sqrt(double(sx.x*sx.x+sy.x*sy.x))));
	sl.y=abs(int(sqrt(double(sx.y*sx.y+sy.y*sy.y))));
	sl.z=abs(int(sqrt(double(sx.z*sx.z+sy.z*sy.z))));
	sl.w=abs(int(sqrt(double(sx.w*sx.w+sy.w*sy.w))));
	
	if(sl.x>85)
		result.x=255;
	else
		result.x=0;
		
	if(sl.y>85)
		result.y=255;
	else
		result.y=0;
		
	if(sl.z>85)
		result.z=255;
	else
		result.z=0;
		
	if(sl.w>85)
		result.w=255;
	else
		result.w=0;
		
	outEdge[global_index]=result;
}

void main()
{
	FILE *ip,*dp;
	char *cur_file;
	int i,j,*image;
	unsigned char *val,*bin_image;
	uchar4 *edge_out;
	
	cur_file="D:\\CUDA_Testing\\dasdsa\\03266.bmp";
	ip=fopen(cur_file,"rb");

	image=(int*)calloc(h*ww,sizeof(int));
	bin_image=(unsigned char*)calloc(h*ww,sizeof(unsigned char));
	val=(unsigned char*)calloc(h*ww,sizeof(unsigned char));
	size_t size=h*ww*sizeof(unsigned char);		
	fseek(ip,54,SEEK_CUR);

	// Getting pixel values of image: START
	for(i=0;i<h;i++)
	{
		for(j=0;j<ww;j++)
		{
			image[i*ww+j]=int(floor((0.3*(int)getc(ip))+(0.59*(int)getc(ip))+(0.11*(int)getc(ip))));			
		}
	}

	for(i=h-1;i>=0;i--)
	{
		for(j=0;j<ww;j++)
		{
			val[(h-1-i)*ww+j]=(unsigned char)image[i*ww+j];
		}
	}
	// Getting pixel values of image: END

	// Sobel Edge Detection on GPU: START
	cudaSetDevice(1);
	cudaMalloc((void **) &edge_out, size);   // Allocate array on device
	
	// allocate array and copy image data
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(8, 0, 0, 0, cudaChannelFormatKindSigned);
    cudaArray* cu_array;
    cudaMallocArray( &cu_array, &channelDesc, ww, h); 
    cudaMemcpyToArray(cu_array, 0, 0, val, size, cudaMemcpyHostToDevice);

    // set texture parameters
    tex.addressMode[0] = cudaAddressModeClamp;
    tex.addressMode[1] = cudaAddressModeClamp;
    tex.filterMode = cudaFilterModePoint;
    tex.normalized = 0;

    // Bind the array to the texture
    cudaBindTextureToArray(tex, cu_array);

	// Define number of threads in a block i.e., block size and number of blocks in a grid i.e., grid size
	dim3 dimBlock(block_size_x,block_size_y);
	dim3 dimGrid(ww/(block_size_x*4),h/block_size_y);
	find_edge<<< dimGrid, dimBlock>>>(edge_out, ww, h);
	cudaMemcpy(bin_image, (unsigned char*)edge_out, size, cudaMemcpyDeviceToHost);	
	cudaFreeArray(cu_array);
	cudaFree(edge_out);
	// Sobel Edge Detection on GPU: END
	
	dp=fopen(".\\edge.pgm","wb");
	fprintf(dp, "P5\n%d %d\n255\n", ww, h);
	for(i=0;i<h;i++)
		for(j=0;j<ww;j++)		
			fprintf(dp,"%c",bin_image[i*ww+j]);		
}